class Willow {
  String axiom = "FX";
  String computedWord;
  String [][] rules = {
    {"F", "FF", "F", "F", "F", "F", "+F-", "-F+"},
    {"X", "+X-", "-X+", "F[++X][--X]", "F[++X][--X]", "F[++X][--X]", "F[++X][FX][--X]", "F[+++X][++X][---X]"},
    {"+", "+", "+", "+", "+", "+", "+", "+"},
    {"-", "-", "-", "-", "-", "-", "-", "-"}, 
    {"[", "[", "[", "[", "[", "[", "[", "["},
    {"]", "]", "]", "]", "]", "]", "]", "]"}, 
  };
  int iterations;
  Boolean simple;
  float size, angle, scale;

  Willow(float size, float angle, float scale, Boolean simple) {
    this.size = size;
    this.angle = angle;
    this.scale = scale;
    this.simple = simple;
    computedWord = axiom;
    iterations = 0;
    colorMode(HSB, 360, 100, 100);
  }
  
  void reset() {
    this.computedWord = this.axiom;
    this.iterations = 0;
  }
  
  void iteration() {
    iterations++;
    char s;
    String newWord = "";
    // loop through computedWord
    for (int i=0; i<computedWord.length(); i++) {
      s = computedWord.charAt(i);
      // for every symbol, check if a rule applies
      for (int j=0; j<rules.length; j++) {
        if (s==rules[j][0].charAt(0)) {
          // replace symbol according to rule
          //newWord = newWord + rules[j][1];
          newWord = newWord + rules[j][(int)random(1, 8)];
          break;
        }
      }
    }
    computedWord = newWord;    
    //println(iterations + ": " + computedWord);
  }
  
  void drawWillow(Turtle t) {
    float strokeWidth = 20;
    float scaledSize = size * pow(scale, iterations);
    // loop through computedWord and draw each symbol
    char s;
    for (int i=0; i<computedWord.length(); i++) {
      s = computedWord.charAt(i);
      if(s == 'F') {
        stroke(22, 29, int(80 - strokeWidth*1.2));
        strokeWeight(strokeWidth);
        t.right(random(-0.5, 0.5));
        t.forward(scaledSize);
      } else if(s == 'X') {
        // draw leaves
        if(this.simple) {
          drawSimple(t);
        } else {
          drawLeaves(t);
        }
      } else if(s == '+') {
        t.right(angle+random(-0.5, 0.5));
      } else if(s == '-') {
        t.left(angle+random(-0.5, 0.5));
      } else if(s == '[') {
        strokeWidth = strokeWidth / 1.5;
        if (strokeWidth <= 1) {
          strokeWidth = 1;
        }
        t.push();
      } else if(s == ']') {
        strokeWidth = strokeWidth * 1.5;
        if (strokeWidth >= 20) {
          strokeWidth = 20;
        }
        t.pop();
      }
    }
  }
  
  void drawSimple(Turtle t) {
    color color1 = #AEDE82;
    color color2 = #7EA25D;

    float t_X = t.getX();
    float t_Y = t.getY();

    strokeWeight(2);
    stroke(color1);
    fill(color1);
    ellipse(t_X, t_Y, 10, 10);
    for(float y = t_Y; y<t_Y+100; y++) {
      float inter = map(y, t_Y, t_Y+100, 0, 1);
      color c = lerpColor(color1, color2, inter);
      stroke(c);
      line(t_X-5, y, t_X+5, y);
    }
    stroke(color2);
    fill(color2);
    ellipse(t_X, t_Y+100, 10, 10);
  }
  
  void drawLeaves(Turtle t) {
    t.push();
    
    strokeWeight(1);
    stroke(136, 39, 100);

    float len = random(50, 100);
    boolean dir = random(0, 1) > 0.5;
    t.setHeading(180);
    float spacing = 10;
    for(int i=0; i<len; i+=spacing) {
      t.push();
      if (dir) {
        t.right(this.angle+random(-1, 1));
      } else {
        t.left(this.angle+random(-1, 1));
      }
      t.forward(2);
      drawLeaf(t, 1);
      t.pop();
      t.forward(spacing);
      dir = !dir;
    }
        
    t.pop();
  }
  
  void drawLeaf(Turtle t, float scale) {
    strokeWeight(1);
    fill(136, 34, 84);
    stroke(136, 39, 68);

    t.push();
    beginShape();
    t.left(15);
    t.forward(5*scale);
    vertex(t.getX(), t.getY());     
    t.right(15);
    t.forward(3*scale);
    vertex(t.getX(), t.getY());   
    t.right(15);
    t.forward(5*scale);
    vertex(t.getX(), t.getY());   
    t.right(150);
    t.forward(5*scale);
    vertex(t.getX(), t.getY());
    t.right(15);
    t.forward(3*scale);
    vertex(t.getX(), t.getY());
    t.right(15);
    t.forward(5*scale);
    vertex(t.getX(), t.getY());
    endShape();
    t.pop();
  }
}
